/* 
	----------- SCLA Team ----------------
				LAPG data struct header
	--------------------------------------
	            		2007.06.02 by hopesjs 
 */
#include "lapg_struct.h"

// Ʈ  Լ
#define BITSPERBYTE     8
#define NBYTES(nbits)   (((nbits) + BITSPERBYTE - 1) / BITSPERBYTE)
#define BIT2BYTE(ibit)  ((ibit) / BITSPERBYTE)
#define BIT2SHIFT(ibit) ((ibit) % BITSPERBYTE)
#define BIT2MASK(ibit)  (1 << BIT2SHIFT(ibit))
#define BYTE2BIT(ibyte) ((ibyte) * BITSPERBYTE)

// LAPG ACKȣ
#define LAPG_ACK_OK				0xAA
#define LAPG_ACK_ERROR		0x55

//   
static unsigned char lapg_reset_command = 0x80;
static unsigned char lapg_exe_command = 0x61;
static unsigned int lapg_send_data_bits_size = 0;							//    ο 
static unsigned char* lapg_send_data_field = NULL;								//    
static unsigned int lapg_receive_data_bits_size = 0;							//  ޴  ο 
static unsigned char* lapg_receive_data_field = NULL;								//  ޴  

// UART   
#define BAUDRATE_UART B115200		

#define _POSIX_SOURCE 1 // POSIX compliant source
#define true		1
#define false		0

static int fd_UART;
static struct termios tio_UART;

/////////////////////////////////////////////////////////////////////////////
//////////////////////// Ʈ  ζ Լ //////////////////////////////
/////////////////////////////////////////////////////////////////////////////

__inline__ void bit_set( unsigned char *flags, int bit )
{
   *flags |= (1U << bit);
}

__inline__ void bit_clr( unsigned char *flags, int bit )
{
   *flags &= ~(1U << bit);
}

__inline__ unsigned char bit_return_value( unsigned char *flags, unsigned char bit_size, int bit)
{
	return ((*flags >> bit) & bit_size);
}

__inline__ void bit_set_value( unsigned char *flags, unsigned char *value , unsigned char bit_size, int bit)
{
	*flags &= ~(bit_size << bit);		//ϴ Ʈ .(  )
	*value &= bit_size;		//  ȿ . 0,1 ǹǷ.
	*flags |= (*value << bit);		// value 
}


/////////////////////////////////////////////////////////////////////////////
////////////////////////// reset command mode setting ///////////////////////
/////////////////////////////////////////////////////////////////////////////

// ¸  Ʈ Ѵ.
// mode : ٲ  (define ) , value :  ( 庰 ũ⿡ ° )
void set_reset_command(int mode, unsigned char value)
{
	switch(mode)
	{
		case LAPG_RESET_CONFIGRATION :
			bit_set_value(&lapg_reset_command, &value, 0x01, 6);		//7° Ʈ 
			break;
		case LAPG_RESET_DCLK :
			bit_set_value(&lapg_reset_command, &value, 0x03, 4);		//6, 5° Ʈ 			
			break;
		case LAPG_RESET_EDGE :
			bit_set_value(&lapg_reset_command, &value, 0x01, 3);		//4° Ʈ 
			break;
		case LAPG_RESET_EXE_MODE :
			bit_set_value(&lapg_reset_command, &value, 0x01, 2);		//3° Ʈ 
			break;
		case LAPG_RESET_PROPAGATION_DELAY :
			bit_set_value(&lapg_reset_command, &value, 0x03, 0);		//2, 1° Ʈ 
			break;
		default:
			lapg_reset_command = 0x80;
			break;
	}
}

// ¸  Ʈ Ѵ.
// mode :  (define )
unsigned char read_reset_command(int mode)
{
	switch(mode)
	{
		case LAPG_RESET_CONFIGRATION :
			return bit_return_value(&lapg_reset_command, 0x01, 6);
			break;
		case LAPG_RESET_DCLK :
			return bit_return_value(&lapg_reset_command, 0x03, 4);
			break;
		case LAPG_RESET_EDGE :
			return bit_return_value(&lapg_reset_command, 0x01, 3);
			break;
		case LAPG_RESET_EXE_MODE :
			return bit_return_value(&lapg_reset_command, 0x01, 2);
			break;
		case LAPG_RESET_PROPAGATION_DELAY :
			return bit_return_value(&lapg_reset_command, 0x03, 0);
			break;
		default:
			return 0;
			break;
	}
}

/////////////////////////////////////////////////////////////////////////////
////////////////////////// execute command mode setting /////////////////////
/////////////////////////////////////////////////////////////////////////////

//   Ʈ Ѵ.
// mode : ٲ  (define ) , value :  ( 庰 ũ⿡ ° )
void set_exe_command(int mode, unsigned char value)
{
	switch(mode)
	{
		case LAPG_EXECUTE_INJECTION :
			bit_set_value(&lapg_exe_command, &value, 0x01, 6);		//7° Ʈ 
			break;
		case LAPG_EXECUTE_SAMPLING :
			bit_set_value(&lapg_exe_command, &value, 0x01, 5);		//6° Ʈ 			
			break;
		case LAPG_EXECUTE_CNT_BASE :
			bit_set_value(&lapg_exe_command, &value, 0x01, 4);		//5° Ʈ 
			break;
		case LAPG_EXECUTE_COUNT :
			bit_set_value(&lapg_exe_command, &value, 0x0f, 0);		//4,3,2,1° Ʈ 
			break;
		default:
			lapg_exe_command = 0x61;
			break;
	}
}

//   Ʈ Ѵ.
// mode :  (define )
unsigned char read_exe_command(int mode)
{
	switch(mode)
	{
		case LAPG_EXECUTE_INJECTION :
			return bit_return_value(&lapg_exe_command, 0x01, 6);
			break;
		case LAPG_EXECUTE_SAMPLING :
			return bit_return_value(&lapg_exe_command, 0x01, 5);
			break;
		case LAPG_EXECUTE_CNT_BASE :
			return bit_return_value(&lapg_exe_command, 0x01, 4);
			break;
		case LAPG_EXECUTE_COUNT :
			return bit_return_value(&lapg_exe_command, 0x0f, 0);
			break;
		default:
			return 0;
			break;
	}
}

/////////////////////////////////////////////////////////////////////////////
//////////////////////////data field's data setting /////////////////////////
/////////////////////////////////////////////////////////////////////////////

// bitŭ  ʵ带 ޸ ҴѴ.
void malloc_data_field(unsigned int send_bits, unsigned int receive_bits)
{
	int i;
	lapg_send_data_bits_size=send_bits;
	lapg_send_data_field=(unsigned char*)malloc(NBYTES(lapg_send_data_bits_size)*sizeof(unsigned char));
	for(i=0;i<NBYTES(lapg_send_data_bits_size);i++)
	{
		lapg_send_data_field[i]=0x00;
	}
	lapg_receive_data_bits_size=receive_bits;
	lapg_receive_data_field=(unsigned char*)malloc(NBYTES(lapg_receive_data_bits_size)*sizeof(unsigned char));
	for(i=0;i<NBYTES(lapg_receive_data_bits_size);i++)
	{
		lapg_receive_data_field[i]=0x00;
	}

}

//  ʵ带 ޸ Ѵ.
void free_data_field()
{
	free(lapg_send_data_field);
	free(lapg_receive_data_field);
	lapg_send_data_bits_size=0;
	lapg_receive_data_bits_size=0;
}

//   ʵ x ġ  value(bit) Ѵ.
void set_send_data_field(unsigned int bit_position, unsigned char value)
{
	bit_set_value((lapg_send_data_field+BIT2BYTE(bit_position)),&value, 0x01, BIT2SHIFT(bit_position));
}

//   ʵ x ġ bit о´.
unsigned char read_send_data_field(unsigned int bit_position)
{
	return bit_return_value((lapg_send_data_field+BIT2BYTE(bit_position)), 0x01, BIT2SHIFT(bit_position));
}

// ޴  ʵ x ġ  value(bit) Ѵ.
void set_receive_data_field(unsigned int bit_position, unsigned char value)
{
	bit_set_value((lapg_receive_data_field+BIT2BYTE(bit_position)),&value, 0x01, BIT2SHIFT(bit_position));
}

// ޴  ʵ x ġ bit о´.
unsigned char read_receive_data_field(unsigned int bit_position)
{
	return bit_return_value((lapg_receive_data_field+BIT2BYTE(bit_position)), 0x01, BIT2SHIFT(bit_position));
}

/////////////////////////////////////////////////////////////////////////////
///////////////////////// serial communication function /////////////////////
/////////////////////////////////////////////////////////////////////////////


// UART Ʈ . (device_uart Ʈ)
void UART_open(char *device_uart)
{
	/*Non-blocking  ø ġ .(readԼ ȣ  ﰢ )*/
	fd_UART = open(device_uart, O_RDWR | O_NOCTTY);// | O_NONBLOCK );//
	if(fd_UART < 0){perror(device_uart); exit(-1);}
	//	printf("success UART Open\n");
		
	/*canonical Է ó  Ʈ*/
	/*BAUDRATE_UART :  ӵ.*/
	/*CS8 : N81*/
	/*CLOCAL : Local connection.    ʴ´. */
//	tcgetattr(fd_UART, &tio_UART);
	tio_UART.c_cflag = BAUDRATE_UART | CS8 | CLOCAL;
	tio_UART.c_iflag = IGNPAR ; /*Parity ִ ڹƮ Ѵ.*/
	tio_UART.c_oflag = 0;
	tio_UART.c_lflag = 0;
	tio_UART.c_cc[VMIN] = 0;
	tio_UART.c_cc[VTIME] = 10;
	tcflush(fd_UART, TCIFLUSH);
	tcsetattr(fd_UART, TCSANOW, &tio_UART);
}

// UART Ʈ ݴ´.
void UART_close(void)
{
	close(fd_UART);
}

// LAPG reset command .
int UART_send_reset_command(void)
{
	unsigned char data = LAPG_ACK_ERROR;

	write(fd_UART, &lapg_reset_command, 1);	
	read(fd_UART, &data, 1);
	if(data == LAPG_ACK_OK)
		return true;
	else
		return false;
}

// LAPG execute command .
int UART_send_execute_command(void)
{
	unsigned char data = LAPG_ACK_ERROR;

	write(fd_UART, &lapg_exe_command, 1);	
	read(fd_UART, &data, 1);
	if(data == LAPG_ACK_OK)
	{
		return true;
	}
	else
		return false;
}

// LAPG send_data field  receive޴´.
void UART_data_field(void)
{
	unsigned char temp=0x00;
	int i;
	if(read_exe_command(LAPG_EXECUTE_INJECTION)==0x01)
	{
		write(fd_UART, lapg_send_data_field, NBYTES(lapg_send_data_bits_size));
	}
	else
	{
		write(fd_UART, &temp, 1);
	}
	if(read_exe_command(LAPG_EXECUTE_SAMPLING)==0x01)
	{
		for(i=0;i<NBYTES(lapg_receive_data_bits_size);i++)
		{
			read(fd_UART, &lapg_receive_data_field[i], 1);	
		}
	}
}

// print send_data_field
void UART_print_lapg_struct(void)
{
	int i;

	printf ("reset_command: %.2x\n", lapg_reset_command);
	printf ("execute_command: %.2x\n", lapg_exe_command);
	printf ("data_field_size : in - %.2d, out - %.2d \n", lapg_send_data_bits_size, lapg_receive_data_bits_size);

	if(lapg_send_data_bits_size!=0)
	{
		printf ("send_data_field : ");
		for(i=0;i<NBYTES(lapg_send_data_bits_size);i++)
		{
			printf ("%.2x ", lapg_send_data_field[i]);
		}
		printf("\n");
	
		printf ("receive_data_field : ");
		for(i=0;i<NBYTES(lapg_receive_data_bits_size);i++)
		{
			printf ("%.2x ", lapg_receive_data_field[i]);
		}
		printf("\n");
	}
}

int UART_get_lapg_bits_size(void)
{
	static unsigned char get_size_command=0xC0;
	unsigned char data[2]={0,0};
	
	write(fd_UART, &get_size_command, 1);
	read(fd_UART, &data[0], 1);
	read(fd_UART, &data[1], 1);
	if(data[0]==0) return false;
	
	malloc_data_field((unsigned int)data[0]*8, (unsigned int)data[1]*8);
	
	return true;
}

// single run
int UART_single_run(void)
{
	int error;
	error = UART_send_reset_command();
	if(error == false) return false;
	lapg_exe_command = 0x61;		// injection, sampling, execute 1
	error = UART_send_execute_command();
	if(error == false) return false;
	UART_data_field();
	return true;
}

// burst run
/*
int UART_burst_run(void)
{
	int error;
	error = UART_send_reset_command();
	if(error == false) return false;
	lapg_exe_command = 0x61;		// injection, sampling, execute 1
	error = UART_send_execute_command();
	if(error == false) return false;
	UART_data_field();
	return true;
}
*/

/////////////////////////////////////////////////////////////
///////////////////// module test code //////////////////////
/////////////////////////////////////////////////////////////
/*
int main()
{
	int i;
	int error;
	UART_open("/dev/ttyS0");
	
	error = UART_get_lapg_bits_size();

	if(error == false) return false;

	lapg_send_data_field[0]=0x32;
	lapg_send_data_field[1]=0x55;
	lapg_send_data_field[2]=0x43;
	lapg_send_data_field[3]=0x35;
	lapg_send_data_field[4]=0x24;
	lapg_send_data_field[5]=0x0D;
	lapg_reset_command=0x98;

	error = UART_single_run();
	if(error == false) return false;
		
	UART_print_lapg_struct();
	
	UART_close();
	

	return 0;
}
*/