library IEEE;
use IEEE.std_logic_1164.ALL;
use IEEE.std_logic_unsigned.ALL;
use IEEE.std_logic_arith.ALL;

package LAPG_PACK is
  
	type STATE is (REC_READY, RECEIVE, COMMAND,  
				INJECTION, PG_DELAY, SAMPLING, TOGGLING, 
				TRANS_READY, TRANSMIT);
	
	constant MAX_BAUD_RATE 			: integer := 1152;
	constant MAX_BIT_CNT 			: integer range 0 to 8 := 8;
	constant MAX_TRANS_CNT			: integer range 0 to 12 := 11;
	constant MAX_CONF_BYTE_NUM		: integer := 2;
	constant ACK_OK				: std_logic_vector(7 downto 0) := "10101010";
	constant ACK_ERROR				: std_logic_vector(7 downto 0) := "01010101";
	function cal_count(cnt_base : std_logic; count_bit : std_logic_vector(3 downto 0)) return integer;	
	
end LAPG_PACK;

package body LAPG_PACK is

	function cal_count(cnt_base : std_logic; count_bit : std_logic_vector(3 downto 0)) return integer is
		variable cnt : integer := 0;
	begin
		if(cnt_base = '0') then
			case count_bit is
				when "0001" => cnt := 1;
				when "0010" => cnt := 4;
				when "0100" => cnt := 16;
				when "1000" => cnt := 64;
				when others => cnt := 0;
			end case;
		else
			case count_bit is
				when "0001" => cnt := 256;
				when "0010" => cnt := 1024;
				when "0100" => cnt := 4096;
				when "1000" => cnt := 16384;
				when others => cnt := 0;
			end case;
		end if;
		return cnt;
	end cal_count;	
end LAPG_PACK;