-- hopesjs computer alu logic
library ieee;
use ieee.std_logic_1164.all;

entity alu is
generic (size: integer := 16);
port(
	dr, ac: in std_logic_vector(size-1 downto 0);
	inpr: in std_logic_vector(7 downto 0);
	cin: in std_logic;	
	fn : in std_logic_vector(2 downto 0);	-- ALU ɾ
	cout: out std_logic;
	q: out std_logic_vector(size-1 downto 0)
);
end alu;

architecture ALU_ARCH of alu is
	component and16 port(
		x,y: in std_logic_vector(size-1 downto 0);
		z: out std_logic_vector(size-1 downto 0)
	);
	end component;

	component fadder16 port(
		x,y: in std_logic_vector(size-1 downto 0);
		s: out std_logic_vector(size-1 downto 0);
		cout: out std_logic
	);
	end component;

	signal caout: std_logic;
	signal out_and, out_add: std_logic_vector(size-1 downto 0);
begin
	mod1: and16 port map (dr, ac, out_and);
	mod2: fadder16 port map(dr, ac, out_add, caout);
	
	q <= out_and when fn="000"						-- AND
		else out_add when fn="001"					-- ADD
		else dr when fn="010"						-- DR Pass
		else "00000000" & inpr when fn="011"		-- INPR Pass
		else (not ac) when fn="100"					-- Compliment AC
		else cin & ac(15 downto 1) when fn="101"	-- shift right
		else ac(14 downto 0) & cin when fn="110"	-- shift left
		else ac when fn="111"				-- AC Pass
		else "0000000000000000";
	cout <= caout when fn="001" else cin;
end ALU_ARCH;
