library IEEE;
use IEEE.std_logic_1164.ALL;
use IEEE.std_logic_unsigned.ALL;
use IEEE.std_logic_arith.ALL;

entity alu_test is
	port(
		CLK : in std_logic;
	  RESET : in std_logic;
		RxD : in std_logic;
		TxD : out std_logic
	);
end alu_test;

architecture Behavioral of alu_test is

component LAPG is
	generic(INPUT_BYTE : integer:=6; OUTPUT_BYTE : integer:=3; OSC : integer:=25);
	port (
		CLK : in std_logic;
		RESET : in std_logic;
		RxD : in std_logic;
		Sampled_Data : in std_logic_vector((OUTPUT_BYTE * 8) - 1 downto 0);
		TxD : out std_logic;
		DCLK : out std_logic;
		Injection_Data	: out std_logic_vector((INPUT_BYTE * 8) - 1 downto 0)
		);
end component;

component alu is
	generic (size: integer := 16);
	port(
		dr, ac: in std_logic_vector(size-1 downto 0);
		inpr: in std_logic_vector(7 downto 0);
		cin: in std_logic;	
		fn : in std_logic_vector(2 downto 0);	-- ALU ɾ
		cout: out std_logic;
		q: out std_logic_vector(size-1 downto 0)
	);
end component;

 	signal input				: std_logic_vector(47 downto 0); -- := "0000000000000000000000000000000000000000000000000000000000000000";
	signal probe				: std_logic_vector(23 downto 0); -- := "0000000000000000000000000000000000000000000000000000000000000000";
	signal DCLK					: std_logic; -- := '0';

begin

U1: LAPG
port map (
	CLK => CLK,			
	RESET	=> RESET,						
	RxD => RxD,		
	TxD => TxD,	
	DCLK => DCLK,			
	Sampled_Data => probe,
	Injection_Data => input
);

U2: alu
port map (
	dr => input(15 downto 0),
	ac => input(31 downto 16),
	inpr => input(39 downto 32),
	cin => input(40),	
	fn => input(43 downto 41),
	cout	=> probe(16),
	q	=> probe(15 downto 0)
);	

end Behavioral;
