library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

entity segment_driver is
	port(clk, start_stop, clear	: in std_logic;
		count_dec	: out std_logic_vector(7 downto 0);
		sel_decode	: out std_logic_vector(7 downto 0));
		
end segment_driver;

architecture arc of segment_driver is
function dis_seg(cnt : integer range 0 to 15) return std_logic_vector is
variable seg_decode : std_logic_vector(7 downto 0);
begin
	case cnt is
		when 0 =>
			seg_decode := "00111111";
			
		when 1 =>
			seg_decode := "00000110";
			
		when 2 =>
			seg_decode := "01011011";
			
		when 3 =>
			seg_decode := "01001111";
			
		when 4 =>
			seg_decode := "01100110";
			
		when 5 =>
			seg_decode := "01101101";
			
		when 6 =>
			seg_decode := "01111101";
			
		when 7 =>
			seg_decode := "00100111";
			
		when 8 =>
			seg_decode := "01111111";
			
		when 9 =>
			seg_decode := "01100111";
			
		when others =>	
			seg_decode := "00000000";
			
	end case;
	
	return (seg_decode);
end dis_seg;

signal msec10, msec10a, msec10b, sec10, sec6, y_out : integer range 0 to 9;
signal min10, min1, hour10, hour1 : integer range 0 to 9;
signal startstop : std_logic;
signal sel_count : std_logic_vector(2 downto 0);

begin
	p_startstop : process(start_stop, clear)
	begin
		if clear = '1' then
			startstop <= '0';
		elsif start_stop'event and start_stop='1' then
			startstop <= not startstop;
		end if;
	end process;
	
	process(clk, clear)
	begin
		if clear='1' then
			msec10 <= 0;
		elsif clk'event and clk='1' then
			if startstop='1' then
				if msec10=9 then
					msec10 <= 0;
				else
					msec10 <= msec10 + 1;
				end if;
			end if;
		end if;
	end process;
	
	
	p_msec1 : process(clk, clear)
	begin
		if clear='1' then
			msec10a <= 0;
		elsif clk'event and clk='1' then
			if startstop='1' then
				if msec10=9 then
					if msec10a=9 then
						msec10a <= 0;
					else
						msec10a <= msec10a + 1;
					end if;
				end if;
			end if;
		end if;
	end process;
	
	p_msec10 : process(clk, clear)
	begin
		if clear='1' then
			msec10b <= 0;
		elsif clk'event and clk='1' then
			if startstop='1' then
				if msec10=9 and msec10a=9 then
					if msec10b=9 then
						msec10b <= 0;
					else
						msec10b <= msec10b + 1;
					end if;
				end if;
			end if;
		end if;
	end process;
	
	p_sec1 : process(clk, clear)
	begin
		if clear='1' then
			sec10 <= 0;
		elsif clk'event and clk='1' then
			if startstop='1' then
				if msec10=9 and msec10a=9 and msec10b=9 then
					if sec10=9 then
						sec10 <= 0;
					else
						sec10 <= sec10 + 1;
					end if;
				end if;
			end if;
		end if;
	end process;
	
	p_sec10 : process(clk, clear)
	begin
		if clear='1' then
			sec6 <= 0;
		elsif clk'event and clk='1' then
			if startstop='1' then
				if msec10=9 and msec10a=9 and msec10b=9 and sec10=9 then
					if sec6=5 then
						sec6 <= 0;
					else
						sec6 <= sec6 + 1;
					end if;
				end if;
			end if;
		end if;
	end process;
	
	p_min1 : process(clk, clear)
	begin
		if clear='1' then
			min1 <= 0;
		elsif clk'event and clk='1' then
			if startstop='1' then
				if msec10=9 and msec10a=9 and msec10b=9 and sec10=9 and sec6=5 then
					if min1=9 then
						min1 <= 0;
					else
						min1 <= min1 + 1;
					end if;
				end if;
			end if;
		end if;
	end process;
	
	p_min10 : process(clk, clear)
	begin
		if clear='1' then
			min10 <= 0;
		elsif clk'event and clk='1' then
			if startstop='1' then
				if msec10=9 and msec10a=9 and msec10b=9 and sec10=9 and sec6=5 and min1=9 then
					if min10=5 then
						min10 <= 0;
					else
						min10 <= min10 + 1;
					end if;
				end if;
			end if;
		end if;
	end process;
	
	p_hour1 : process(clk, clear)
	begin
		if clear='1' then
			hour1 <= 0;
		elsif clk'event and clk='1' then
			if startstop='1' then
				if msec10=9 and msec10a=9 and msec10b=9 and sec10=9 and sec6=5 and min1=9 and min10=5 then
					if hour1=9 then
						hour1 <= 0;
					else
						hour1 <= hour1 + 1;
					end if;
				end if;
			end if;
		end if;
	end process;
	
	p_hour10 : process(clk, clear)
	begin
		if clear='1' then
			hour10 <= 0;
		elsif clk'event and clk='1' then
			if startstop='1' then
				if msec10=9 and msec10a=9 and msec10b=9 and sec10=9 and sec6=5 and min1=9 and min10=5 and hour1=9 then
					if hour10=5 then
						hour10 <= 0;
					else
						hour10 <= min10 + 1;
					end if;
				end if;
			end if;
		end if;
	end process;
		
	sel_counter : process(clk, clear)
	begin
		if clear='1' then
			sel_count <= "000";
		elsif clk'event and clk='1' then
			if sel_count="111" then
				sel_count <= "000";
			else
				sel_count <= sel_count + 1;
			end if;
		end if;
	end process;
	
	process(sel_count, msec10a, msec10b, sec10, sec6, min1, min10, hour1, hour10)
	begin
		case sel_count is
			when "000" =>
				y_out <= msec10a;
			when "001" =>
				y_out <= msec10b;
			when "010" =>
				y_out <= sec10;
			when "011" =>
				y_out <= sec6;
			when "100" =>
				y_out <= min1;
			when "101" =>
				y_out <= min10;
			when "110" =>
				y_out <= hour1;
			when "111" =>
				y_out <= hour10;
			when others =>
				null;
		end case;
	end process;
	
	count_dec <= dis_seg(y_out);
	
	process(sel_count)
	begin
		case sel_count is
			when "000" =>
				sel_decode <= "11111110";
				
			when "001" =>
				sel_decode <= "11111101";
				
			when "010" =>
				sel_decode <= "11111011";
				
			when "011" =>
				sel_decode <= "11110111";
				
			when "100" =>
				sel_decode <= "11101111";
			
			when "101" =>
				sel_decode <= "11011111";
			
			when "110" =>
				sel_decode <= "10111111";
				
			when "111" =>
				sel_decode <= "01111111";
			
			when others =>
				null;
		end case;
	end process;
end arc;
