/*
 * Copyright (c) Tony Bybell 2005.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

#define DEBUG_MALLOC

#include <config.h>
#include "vcd_saver.h"
#include "helpers/lxt_write.h"
#include "ghw.h"
#include <time.h>
#include "logicanalyzer.h"
#include "lapg_struct.h"
#include "treesearch.h"
#include <glib.h>

static FILE *f = NULL; /* file handle for output file */

/*
 * unconvert trace data back to VCD representation...use strict mode for LXT
 */
static unsigned char analyzer_demang(int strict, unsigned char ch)
{
if(!strict)
	{
	if(ch < AN_COUNT)
		{
		return(AN_STR[ch]);
		}
		else
		{
		return(ch);
		}
	}
else
	{
	if(ch < AN_COUNT)
		{
		return(AN_STR4ST[ch]);
		}
		else
		{
		return(ch);
		}
	}
}


/*
 * generate a vcd identifier for a given facindx
 */
static char *vcdid(int value)
{
static char buf[16];
int i;
                                         
for(i=0;i<15;i++)
        {
        buf[i]=(char)((value%94)+33); /* for range 33..126 */
        value=value/94;
        if(!value) {buf[i+1]=0; break;}
        }
                    
return(buf);
}

static char *vcd_truncate_bitvec(char *s)
{
char l, r;

r=*s;
if(r=='1')
        {
        return s;
        }
        else
        {
        s++;
        }

for(;;s++)
        {
        l=r; r=*s;
        if(!r) return (s-1);

        if(l!=r)
                {
                return(((l=='0')&&(r=='1'))?s:s-1);
                }
        }
}


/************************ splay ************************/

/*
 * integer splay
 */
typedef struct vcdsav_tree_node vcdsav_Tree;
struct vcdsav_tree_node {
    vcdsav_Tree * left, * right;
    nptr item;
    int val;
    hptr hist;
    int len;

    union
	{
	void *p;
	long l;
	int i;
	} handle;	/* future expansion for adding other writers that need pnt/handle info */

    unsigned char flags;
};


static long vcdsav_cmp_l(void *i, void *j)
{
long il = (long)i, jl = (long)j;
return(il - jl);
}

static vcdsav_Tree * vcdsav_splay (void *i, vcdsav_Tree * t) {
/* Simple top down splay, not requiring i to be in the tree t.  */
/* What it does is described above.                             */
    vcdsav_Tree N, *l, *r, *y;
    int dir;

    if (t == NULL) return t;
    N.left = N.right = NULL;
    l = r = &N;

    for (;;) {
	dir = vcdsav_cmp_l(i, t->item);
	if (dir < 0) {
	    if (t->left == NULL) break;
	    if (vcdsav_cmp_l(i, t->left->item)<0) {
		y = t->left;                           /* rotate right */
		t->left = y->right;
		y->right = t;
		t = y;
		if (t->left == NULL) break;
	    }
	    r->left = t;                               /* link right */
	    r = t;
	    t = t->left;
	} else if (dir > 0) {
	    if (t->right == NULL) break;
	    if (vcdsav_cmp_l(i, t->right->item)>0) {
		y = t->right;                          /* rotate left */
		t->right = y->left;
		y->left = t;
		t = y;
		if (t->right == NULL) break;
	    }
	    l->right = t;                              /* link left */
	    l = t;
	    t = t->right;
	} else {
	    break;
	}
    }
    l->right = t->left;                                /* assemble */
    r->left = t->right;
    t->left = N.right;
    t->right = N.left;
    return t;
}


static vcdsav_Tree * vcdsav_insert(void *i, vcdsav_Tree * t, int val, unsigned char flags, hptr h) {
/* Insert i into the tree t, unless it's already there.    */
/* Return a pointer to the resulting tree.                 */
    vcdsav_Tree * n;
    int dir;
    
    n = (vcdsav_Tree *) calloc_2(1, sizeof (vcdsav_Tree));
    if (n == NULL) {
	fprintf(stderr, "vcdsav_insert: ran out of memory, exiting.\n");
	exit(255);
    }
    n->item = i;
    n->val = val;
    n->flags = flags;
    n->hist = h;
    if (t == NULL) {
	n->left = n->right = NULL;
	return n;
    }
    
    t = vcdsav_splay(i,t);
    dir = vcdsav_cmp_l(i,t->item);
    if (dir<0) {
	n->left = t->left;
	n->right = t;
	t->left = NULL;
	return n;
    } else if (dir>0) {
	n->right = t->right;
	n->left = t;
	t->right = NULL;
	return n;
    } else { // We get here if it's already in the tree //
             // Don't add it again                      //
	free_2(n);
	return t;
    }

}

static vcdsav_Tree * vcdsav_insert2(void *i, vcdsav_Tree * t, int val, unsigned char flags, hptr h) {
/* Insert i into the tree t, unless it's already there.    */
/* Return a pointer to the resulting tree.                 */
    vcdsav_Tree * n;
    int dir;
    
    n = (vcdsav_Tree *) calloc_2(1, sizeof (vcdsav_Tree));
    if (n == NULL) {
	fprintf(stderr, "vcdsav_insert: ran out of memory, exiting.\n");
	exit(255);
    }
    n->item = i;
    n->val = val;
    n->flags = flags;
    n->hist = h;
    if (t == NULL) {
	n->left = n->right = NULL;
	return n;
    }
    
//    t = vcdsav_splay(i,t);			//  
//    dir = vcdsav_cmp_l(i,t->item);		// 񱳿    
//    if (dir<0) {
//	n->right = t->right;
//	n->left = t;
//	t->right = NULL;
//	return n;
 //   } else if (dir>0) {
	n->right = t->right;
	n->left = t;
	t->right = NULL;
	return n;
//    } else { // We get here if it's already in the tree //
//             // Don't add it again                      //
//	free_2(n);
//	return t;
//    }

}

/************************ heap ************************/

static int hpcmp(vcdsav_Tree *hp1, vcdsav_Tree *hp2)
{
hptr n1 = hp1->hist;
hptr n2 = hp2->hist;
TimeType t1, t2;

if(n1)  t1 = n1->time; else t1 = MAX_HISTENT_TIME;
if(n2)  t2 = n2->time; else t2 = MAX_HISTENT_TIME;

if(t1 == t2)
	{
	return(0);
	}
else
if(t1 > t2)
	{
	return(-1);
	}
else
	{
	return(1);
	}
}


void recurse_build(vcdsav_Tree *vt, vcdsav_Tree ***hp)
{
if(vt->left) recurse_build(vt->left, hp);

**hp = vt;
*hp = (*hp) + 1;

if(vt->right) recurse_build(vt->right, hp);
}


/* 
 * heapify algorithm...used to grab the next value change
 */
static vcdsav_Tree **hpout = NULL;
static void heapify2(int i, int heap_size,vcdsav_Tree **hp)
{
int l, r;
unsigned int largest;
vcdsav_Tree *t;
int maxele=heap_size/2-1;	/* points to where heapswaps don't matter anymore */
for(;;)
        {
        l=2*i+1;
        r=l+1;
                         
        if((l<heap_size)&&(hpcmp(hp[l],hp[i])>0))
                {
                largest=l;
                }   
                else
                {
                largest=i;
                }
        if((r<heap_size)&&(hpcmp(hp[r],hp[largest])>0))
                {
                largest=r;
                }
        
        if(i!=largest)
                {
                t=hp[i];
                hp[i]=hp[largest];
                hp[largest]=t;
                
                if(largest<=maxele)
                        {
                        i=largest;
                        }
                        else
                        {
                        break;
                        } 
                }   
                else
                {
                break;
                }
        }
}


static vcdsav_Tree **hp = NULL;
static void heapify(int i, int heap_size)
{
int l, r;
unsigned int largest;
vcdsav_Tree *t;
int maxele=heap_size/2-1;	/* points to where heapswaps don't matter anymore */
                
for(;;)
        {
        l=2*i+1;
        r=l+1;
                         
        if((l<heap_size)&&(hpcmp(hp[l],hp[i])>0))
                {
                largest=l;
                }   
                else
                {
                largest=i;
                }
        if((r<heap_size)&&(hpcmp(hp[r],hp[largest])>0))
                {
                largest=r;
                }
        
        if(i!=largest)
                {
                t=hp[i];
                hp[i]=hp[largest];
                hp[largest]=t;
                
                if(largest<=maxele)
                        {
                        i=largest;
                        }
                        else
                        {
                        break;
                        } 
                }   
                else
                {
                break;
                }
        }
}

/* ٷ ؿ ϸ SERIAL Լ        ϱ */
//#define SERIAL(x) 
//#define NOSERIAL(x) x
#define SERIAL(x) x
#define TESTNOSERIAL(x) 
int serial_nodes_to_export(const char *fname)
{

Trptr t = traces.first;
int nodecnt = 0,nodecntout = 0,nodecntclk = 0;
vcdsav_Tree *vt = NULL,*vtout = NULL,*vtclk = NULL;
vcdsav_Tree **hp_clone = hp;
vcdsav_Tree **hpout_clone = hpout; //for out

nptr n;
ExtNode *e;
int msi, lsi;
int i,j;
int iBitIndex = 0,iBitIndexOut = 0;
unsigned char *lapg_receive_data_field_cur = NULL;

TimeType prevtime = LLDescriptor(-1);
TimeType tCyclePeriod = LLDescriptor(-1);

time_t walltime;
struct strace *st = NULL;
int strace_append = 0;
int max_len = 1;
int max_lenout = 1;
char *row_data = NULL;
struct lt_trace *lt = NULL;

DATAINDEX *dataIndex;
DATAINDEX *dataIndexOut;

g_message("serial_nodes_to_export\n");
errno = 0;


f = fopen(fname, "wb");
if(!f)
	{
	return(VCDSAV_FILE_ERROR);
	}

//vcd tree ΰ  input ÷ΰ Output ÷ ΰ
while(t)
	{
	if(!t->vector)
		{
		if(t->n.nd) 
			{
			n = t->n.nd;
			if(n->expansion) n = n->expansion->parent;
			e = n->ext;
			if(e) { msi = e->msi; lsi = e->lsi; 
				
			} else { msi = lsi = -1; }
			vt = vcdsav_splay(n, vt);
			if(!vt || vt->item != n)
				{
				unsigned char flags = 0;

				if(n->head.next)
				if(n->head.next->next)
					{
					flags = n->head.next->next->flags;
					}

				//input = 1  ooutput = 0
				if(t->inoutflag == ACTION_INPUT)
					{
						vt = vcdsav_insert2(n, vt, ++nodecnt, flags, &n->head);
					}
				else if(t->inoutflag == ACTION_OUTPUT){
						vtout = vcdsav_insert2(n, vtout, ++nodecntout, flags, &n->head);
					}
				else if(t->inoutflag == ACTION_CLK){
						vt = vcdsav_insert2(n, vt, ++nodecnt , flags, &n->head);
						vtclk = vcdsav_insert2(n, vtclk, ++nodecntclk, flags, &n->head);
					}
				}
			}
		}	
		else
		{
		bvptr b = t->n.vec;
		if(b)
			{
			bptr bt = b->bits;
			if(bt)
				{
				for(i=bt->nbits-1;i>=0;i--)		//  
					{
					if(bt->nodes[i]) 
						{
						n = bt->nodes[i];
	
						if(n->expansion) n = n->expansion->parent;
						e = n->ext;
						if(e) { msi = e->msi; lsi = e->lsi; } else { msi = lsi = -1; }

						if(!vt || vt->item != n)
							{
							unsigned char flags = 0;

							if(n->head.next)
							if(n->head.next->next)
								{
								flags = n->head.next->next->flags;
								}
							//vcdsav_insert2 == created
							//vt = vcdsav_insert2(n, vt, ++nodecnt, flags, &n->head);
							//input = 1  ooutput = 2
							if(t->inoutflag == ACTION_INPUT)
								{
									vt = vcdsav_insert2(n, vt, ++nodecnt, flags, &n->head);
								}
							else if(t->inoutflag == ACTION_OUTPUT){
									vtout = vcdsav_insert2(n, vtout, ++nodecntout, flags, &n->head);
								}
							else if(t->inoutflag == ACTION_CLK){
									//vt = vcdsav_insert2(n, vt, ++nodecnt, flags, &n->head); //why?
									vtclk = vcdsav_insert2(n, vtclk, ++nodecntclk, flags, &n->head);
								}
							
							//   ʴ 2°  Լ ̿
							}							
						}
					}				
				}
			}
		}

	if(!strace_append)
		{
		t=t->t_next;
		if(!t) 
			{
	                if(shadow_straces)
	                        {
	                        swap_strace_contexts();
				st = straces;

				strace_append = 1;
				t = st ? st->trace : NULL;
				}
			}
		}
		else
		{
		st = st->next;
		t = st ? st->trace : NULL;
		}
	}

g_message("nodecntclk = %d, nodecnt = %d, nodecntout = %d",nodecntclk,nodecnt,nodecntout);

if(strace_append) swap_strace_contexts();

/*it have input but may not need output? */
/*
if(!nodecnt || !nodecntout) {
	return(VCDSAV_EMPTY);
}
*/
if(nodecntclk > 1) 
	{
	return(VCDSAV_TOOMANY_CLK);
}

/* write out netnames here ... */
hp_clone = hp = calloc_2(nodecnt, sizeof(vcdsav_Tree *));
hpout_clone  = hpout = calloc_2(nodecntout, sizeof(vcdsav_Tree *));

recurse_build(vt, &hp_clone);
recurse_build(vtout, &hpout_clone);

dataIndex = (DATAINDEX*)malloc_2(sizeof(DATAINDEX)*nodecnt);
//data input's bit index
for(i=0;i<nodecnt;i++)
	{
	char *netname = output_hier(hp[i]->item->nname);
	if(vtclk && hp[i]->val == vtclk->val){
		continue;
		}
	if(hp[i]->flags & (HIST_REAL|HIST_STRING))
		{
		  //fprintf(f, "$var real 1 %s %s $end\n", vcdid(hp[i]->val), netname);
			dataIndex[i].iBitIndex = iBitIndex;
			dataIndex[i].iSize = 1;
			//16 buffer vcdid Լ  
			strncpy(dataIndex[i].pcVcdid,vcdid(hp[i]->val),16);
			iBitIndex += 1;
			g_message("%d %d %s",dataIndex[i].iBitIndex,dataIndex[i].iSize,dataIndex[i].pcVcdid);	
		}
	else {
		int msi = -1, lsi = -1;

		if(hp[i]->item->ext)
			{
			msi = hp[i]->item->ext->msi;
			lsi = hp[i]->item->ext->lsi;
			}
		
		if(msi==lsi)
			{
				//fprintf(f, "$var wire 1 %s %s $end\n", vcdid(hp[i]->val), netname);
				dataIndex[i].iBitIndex = iBitIndex;
				dataIndex[i].iSize = 1;
				//16 buffer vcdid Լ  
				strncpy(dataIndex[i].pcVcdid,vcdid(hp[i]->val),16);
				iBitIndex += 1;
				g_message("%d %d %s",dataIndex[i].iBitIndex,dataIndex[i].iSize,dataIndex[i].pcVcdid);
			}
			else
			{
				int len = (msi < lsi) ? (lsi - msi + 1) : (msi - lsi + 1);
				//fprintf(f, "$var wire %d %s %s $end\n", len, vcdid(hp[i]->val), netname);
				dataIndex[i].iBitIndex = iBitIndex;
				dataIndex[i].iSize = len;
				//16 buffer vcdid Լ  
				strncpy(dataIndex[i].pcVcdid,vcdid(hp[i]->val),16);
				iBitIndex += len;
				g_message("%d %d %s",dataIndex[i].iBitIndex,dataIndex[i].iSize,dataIndex[i].pcVcdid);
					
				hp[i]->len = len;
				if(len > max_len) max_len = len;
			}			
		}
	}

row_data = malloc_2(max_len + 1);

dataIndexOut = (DATAINDEX*)malloc_2(sizeof(DATAINDEX)*nodecntout);
//data output index
for(i=0;i<nodecntout;i++)
	{	
	if(hpout[i]->flags & (HIST_REAL|HIST_STRING))
		{
			dataIndexOut[i].iBitIndex = iBitIndexOut;
			dataIndexOut[i].iSize = 1;
			//16 buffer vcdid Լ  
			strncpy(dataIndexOut[i].pcVcdid,vcdid(hpout[i]->val),16);
			iBitIndexOut += 1;
			g_message("%d %d %s",dataIndexOut[i].iBitIndex,dataIndexOut[i].iSize,hpout[i]->item->nname);	
		}
	else {
		int msi = -1, lsi = -1;

		if(hpout[i]->item->ext)
			{
			msi = hpout[i]->item->ext->msi;
			lsi = hpout[i]->item->ext->lsi;
			}
		
		if(msi==lsi)
			{
				dataIndexOut[i].iBitIndex = iBitIndexOut;
				dataIndexOut[i].iSize = 1;
				//16 buffer vcdid Լ  
				strncpy(dataIndexOut[i].pcVcdid,vcdid(hpout[i]->val),16);
				iBitIndexOut += 1;
				g_message("1 %d %d %s",dataIndexOut[i].iBitIndex,dataIndexOut[i].iSize,hpout[i]->item->nname);
			}
			else
			{
				int len = (msi < lsi) ? (lsi - msi + 1) : (msi - lsi + 1);
				dataIndexOut[i].iBitIndex = iBitIndexOut;
				dataIndexOut[i].iSize = len;
				//16 buffer vcdid Լ  
				strncpy(dataIndexOut[i].pcVcdid,vcdid(hpout[i]->val),16);
				iBitIndexOut += len;
				g_message("2 %d %d %s",dataIndexOut[i].iBitIndex,dataIndexOut[i].iSize,hpout[i]->item->nname);		
				hpout[i]->len = len;	
				if(len > max_lenout) max_lenout = len;
			}			
		}

	}

//for(i=0;i<nodecntout;i++)
//{
//	g_message("hpout[%d] = %s , dataIndexOut[%d].iBitIndex = %d",i,hpout[i]->item->nname,i, dataIndexOut[i].iBitIndex);
//}

/*
 argument 1 : send data field size
 argument 2 : recive data field size
*/
SERIAL(
	int res=0;
	res=UART_get_lapg_bits_size(NBYTES(iBitIndex), NBYTES(iBitIndexOut));
	switch(res)
	{
		case UART_ERROR:
			g_message("########## lapg_uart_error ##########");
			return VCDSAV_UART_ERROR;
		case INPUT_ERROR:
			g_message("########## lapg_input_error ##########");
			return VCDSAV_UART_ERROR;
		case OUTPUT_ERROR:
			g_message("########## lapg_output_error ##########");
			return VCDSAV_UART_ERROR;
	}

	res = UART_send_reset_command();
	if(res != 1)
	{
		g_message("########## reset_command error ##########");
		return VCDSAV_UART_ERROR;
	}
);

//for test
TESTNOSERIAL(malloc_data_field(iBitIndex, iBitIndexOut));
lapg_receive_data_field_cur = malloc_2( (iBitIndexOut + 7)  / 8);
TESTNOSERIAL(memset(lapg_receive_data_field_cur,'\0',sizeof(char) * (iBitIndexOut + 7)  / 8));

//g_message("%d",(iBitIndexOut + 7)  / 8);
output_hier("");
free_hier();

//fprintf(f, "$enddefinitions $end\n");
//fprintf(f, "$dumpvars\n");

// if first node for define clock time period
if(vtclk){
	hptr next;
	eptr e=ExpandNode(vtclk->item);
	TimeType tPrevTime , tCurTime;	
	if(e){
		next = &e->narray[0]->head;
		free_2(e->narray);
		free_2(e);
		}
	else {
		next = &vtclk->item->head;
		}
	while(next)
		{
		if(next->time >= LLDescriptor(0))
			{
				tPrevTime = next->time;
				// determine start level is
				if(analyzer_demang(WAVE_EXPORT_VCD,next->v.h_val) == '0')
					{
						set_reset_command(LAPG_RESET_DCLK, 2);
					}
				else if(analyzer_demang(WAVE_EXPORT_VCD,next->v.h_val) == '1'){
						set_reset_command(LAPG_RESET_DCLK, 1);
					}
				if(next->next && next->next->time < max_time){						
						tCurTime = next->next->time; 	
						tCyclePeriod = tCurTime - tPrevTime;
					}
				break;
			}
		next = next->next;
		}			
}
g_message("tCyclePeriod = #"TTFormat,tCyclePeriod);
if(tCyclePeriod ==  LLDescriptor(-1)){
	tCyclePeriod = 1;
	g_message("Dosen't have clk period");
	return(VCDSAV_PERIOD_EMPTY);
}


for(i=(nodecnt/2-1);i>0;i--)        /* build nodes into having heap property */
        {
        heapify(i,nodecnt);  
 }

for(;;)
	{
	heapify(0, nodecnt);
	if(!hp[0]->hist) break;
	if(hp[0]->hist->time > max_time) break;
	
	if((hp[0]->hist->time != prevtime) && (hp[0]->hist->time >= LLDescriptor(0)))
		{
		
		TimeType tnorm = hp[0]->hist->time;
		if(time_scale != 1)
			{
			tnorm /= time_scale;
			
			}

		if(tnorm % tCyclePeriod == 0){
			g_message("#"TTFormat, tnorm);
			
			SERIAL(
				int res=0;
				res=UART_single_run(); 
				// res=UART_run(); //not use yet
				
				if(res != 1)
				{
					g_warning("########## serial time out ##########");
					break;
				}
				);
			
			TESTNOSERIAL(
				int i;
				for(i=0;i<NBYTES(iBitIndexOut);i++)
				{
					lapg_receive_data_field[i]=0;
				}
			);				
			memcpy(lapg_receive_data_field_cur,lapg_receive_data_field,NBYTES(iBitIndexOut));
											TESTNOSERIAL(
												unsigned char value;
												if(((int)prevtime)%100 ==0)
												{
													int aa;
													int bb;
													aa=((int)prevtime)/100;
													bb=2;
													if(aa==bb)
													{
														value=0x01;
														bit_set_value((lapg_receive_data_field_cur+BIT2BYTE(bb%8)),&value, 0x01, BIT2SHIFT(bb%8));	
													}
													bb=13;
													if(aa==bb)
													{
														value=0x01;
														bit_set_value((lapg_receive_data_field_cur+BIT2BYTE(bb%8)),&value, 0x01, BIT2SHIFT(bb%8));	
													}
													bb=15;
													if(aa==bb)
													{
														value=0x01;
														bit_set_value((lapg_receive_data_field_cur+BIT2BYTE(bb%8)),&value, 0x01, BIT2SHIFT(bb%8));	
													}
												}		
											);				

			TimeType tPrevtimeForCycle = prevtime - (prevtime % tCyclePeriod); 
			
			if(hp[0]->hist->time > LLDescriptor(0))
			{
				for(i=0;i<nodecntout;i++)
				{
					unsigned char new_h_val;
					while(hpout[i]->hist->next->time < hp[0]->hist->time)
					{
						hpout[i]->hist=hpout[i]->hist->next;
					}
				
				//	g_message("hpout[%d] = %s , dataIndexOut[%d].iBitIndex = %d",i,hpout[i]->item->nname,i, dataIndexOut[i].iBitIndex);

					if(hpout[i]->hist->next->time > hp[0]->hist->time)
					{
						struct HistEnt *he;
			                     he=histent_calloc();
					       he->time=hp[0]->hist->time;
						he->v.h_val= hpout[i]->hist->v.h_val;
						he->next=hpout[i]->hist->next;
						hpout[i]->hist->next = he;	
					//	SERIAL(UART_print_lapg_struct());
					}
					hpout[i]->hist=hpout[i]->hist->next; 
					//g_message("tPrevtimeForCycle = "TTFormat",hpout[i]->hist->time= "TTFormat,tPrevtimeForCycle,hpout[i]->hist->time);
					//g_message("vcd = %c,output = %d",analyzer_demang(WAVE_EXPORT_VCD,hpout[i]->hist->v.h_val), read_data_field(dataIndexOut[i].iBitIndex,lapg_receive_data_field_cur));									

					switch(hpout[i]->hist->v.h_val)
					{
						case AN_0:
						case AN_RSV9:
							if(read_data_field(dataIndexOut[i].iBitIndex,lapg_receive_data_field_cur)==0)
								hpout[i]->hist->v.h_val=AN_0;
							else
								hpout[i]->hist->v.h_val=AN_RSV9;
						break;
						case AN_1:
						case AN_RSVA:
							if(read_data_field(dataIndexOut[i].iBitIndex,lapg_receive_data_field_cur)==0)
								hpout[i]->hist->v.h_val=AN_RSVA;
							else
								hpout[i]->hist->v.h_val=AN_1;
						break;
						default:
							hpout[i]->hist->v.h_val=AN_X;
						break;
					}
				}
			}
			wavearea_configure_event(wavearea, NULL);
			}
		prevtime = hp[0]->hist->time;	
	}

	if(hp[0]->hist->time >= LLDescriptor(0))
		{
		if(hp[0]->flags & (HIST_REAL|HIST_STRING))
			{
			if(hp[0]->flags & HIST_STRING)
				{
				char *vec = hp[0]->hist->v.h_vector ? hp[0]->hist->v.h_vector : "UNDEF";
				fprintf(f, "s%s %s\n", vec, vcdid(hp[0]->val));
				}
				else
				{
				double *d = (double *)hp[0]->hist->v.h_vector;
                double value;

				if(!d)
				{
                                sscanf("NaN", "%lg", &value);
				}
				else
				{
					value = *d;
				}
				fprintf(f, "r%.16g %s\n", value, vcdid(hp[0]->val));	
			
				}	
			}
		else
		if(hp[0]->len)
			{
			if(hp[0]->hist->v.h_vector)
				{
				for(i=0;i<hp[0]->len;i++)
					{
					row_data[i] = analyzer_demang(WAVE_EXPORT_VCD, hp[0]->hist->v.h_vector[i]);
					}				
				}
				else
				{
				for(i=0;i<hp[0]->len;i++)
					{
					row_data[i] = 'x';
					}				
				}
				row_data[i] = 0;
			
				fprintf(f, "b%s %s\n", vcd_truncate_bitvec(row_data), vcdid(hp[0]->val));
				//bit is 1 or 0, size is one
				for(i=0;i<nodecnt;i++)
					{					
						if(strcmp(dataIndex[i].pcVcdid,vcdid(hp[0]->val)) == 0 ){
							for(j=0;j<dataIndex[i].iSize;j++)
								{
								switch(row_data[dataIndex[i].iSize - (j+1)])
									{
									case '0' :
										SERIAL(set_send_data_field(dataIndex[i].iBitIndex + j,(unsigned char)0));
										break;
									case '1' :
										SERIAL(set_send_data_field(dataIndex[i].iBitIndex + j,(unsigned char)1));
										break;
									}
								g_message("IN:i = %d, position = %d, data = %c, vcdid = %s , size = %d", i, dataIndex[i].iBitIndex + j, row_data[dataIndex[i].iSize - (j+1)],dataIndex[i].pcVcdid,dataIndex[i].iSize);
								}
							break;
							}
					}
			}
		else
			{					
				fprintf(f, "%c%s\n", analyzer_demang(WAVE_EXPORT_VCD, hp[0]->hist->v.h_val), vcdid(hp[0]->val));
					
				/*bit is 1 or 0, size is one */
				for(i=0;i<nodecnt;i++)
					{
						
						if( strcmp(dataIndex[i].pcVcdid,vcdid(hp[0]->val)) == 0 ){
								switch(analyzer_demang(WAVE_EXPORT_VCD, hp[0]->hist->v.h_val))
									{
									case '0' :
										SERIAL(set_send_data_field(dataIndex[i].iBitIndex,(unsigned char)0));
										break;
									case '1' :
										SERIAL(set_send_data_field(dataIndex[i].iBitIndex,(unsigned char)1));
										break;									
									}
								g_message("IN:i = %d, position = %d, data = %c, vcdid1 = %s, vcdid2 = %s, size = %d",
										  i, 
										  dataIndex[i].iBitIndex, 
										  analyzer_demang(WAVE_EXPORT_VCD, hp[0]->hist->v.h_val),
										  dataIndex[i].pcVcdid,vcdid(hp[0]->val),
										  dataIndex[i].iSize);
								break;
							}
					}				
			}
		}
	
	hp[0]->hist = hp[0]->hist->next;
	}

if(prevtime < max_time)
	{
	
		fprintf(f, "#"TTFormat"\n", max_time / time_scale);

	}
/*
eptr e=ExpandNode(t->n.nd);
					int i;
					if(!e)
						{
						if(t->n.nd->expansion) t->n.nd->expansion->refcnt++;
						AddNode(t->n.nd,NULL);
						}
						else
						{
						for(i=0;i<e->width;i++)
							{
							AddNode(e->narray[i], NULL);						
							}
						free_2(e->narray);
						free_2(e);
						}
						*/
//if node array
/*
for(i=0;i<nodecntout;i++)
	{
	hptr next;
	eptr e=ExpandNode(hpout[i]->item);
	if(e){
		for(j=0;j<e->width;j++)
		{
			next = &e->narray[j]->head;
			while(next)
				{
				g_message("NODE VECTOR: #"TTFormat" %c",next->time,analyzer_demang(WAVE_EXPORT_VCD, next->v.h_val));
				next->v.h_val = 0;
				next = next->next;
				signalarea_configure_event(signalarea, NULL);
				wavearea_configure_event(wavearea, NULL);
				}
		}
		free_2(e->narray);
		free_2(e);
		}
	else {
		
		if(hpout[i]->item->expansion) hpout[i]->item->expansion->refcnt++;
		
		next = &hpout[i]->item->head;
		while(next)
				{
				g_message("NODE:#"TTFormat" %c",next->time,analyzer_demang(WAVE_EXPORT_VCD, next->v.h_val));
				next->v.h_val = 0;
				next = next->next;
				
				}
		}
	}

MaxSignalLength();
signalarea_configure_event(signalarea, NULL);
wavearea_configure_event(wavearea, NULL);
	*/			
//showchange("Show-Change", t, NULL);
for(i=0;i<nodecnt;i++)
	{
	free_2(hp[i]);
	}

free_2(hp); hp = NULL;

for(i=0;i<nodecntout;i++)
	{
	free_2(hpout[i]);
	}

free_2(hpout); hpout = NULL;

free_2(row_data); row_data = NULL;
SERIAL(free_data_field());
free_2(dataIndex); dataIndex = NULL;
free_2(lapg_receive_data_field_cur);


fclose(f); f = NULL;

return(VCDSAV_OK);
}

/*
 * mainline
 */ 
int save_nodes_to_export(const char *fname, int export_typ)
{
Trptr t = traces.first;
int nodecnt = 0;
vcdsav_Tree *vt = NULL;
vcdsav_Tree **hp_clone = hp;
nptr n;
ExtNode *e;
int msi, lsi;
int i;
TimeType prevtime = LLDescriptor(-1);
time_t walltime;
struct strace *st = NULL;
int strace_append = 0;
int max_len = 1;
char *row_data = NULL;
struct lt_trace *lt = NULL;
int lxt = (export_typ == WAVE_EXPORT_LXT);

errno = 0;
if(lxt)
	{
	lt = lt_init(fname);
	if(!lt)
		{
		return(VCDSAV_FILE_ERROR);
		}
	}
	else
	{
	f = fopen(fname, "wb");
	if(!f)
		{
		return(VCDSAV_FILE_ERROR);
		}
	}

while(t)
	{
	if(!t->vector)
		{
		if(t->n.nd) 
			{
			n = t->n.nd;
			if(n->expansion) n = n->expansion->parent;
			e = n->ext;
			if(e) { msi = e->msi; lsi = e->lsi; } else { msi = lsi = -1; }
			vt = vcdsav_splay(n, vt);
			if(!vt || vt->item != n)
				{
				unsigned char flags = 0;

				if(n->head.next)
				if(n->head.next->next)
					{
					flags = n->head.next->next->flags;
					}

				vt = vcdsav_insert(n, vt, ++nodecnt, flags, &n->head);
				}
			}
		}	
		else
		{
		bvptr b = t->n.vec;
		if(b)
			{
			bptr bt = b->bits;
			if(bt)
				{
				for(i=0;i<bt->nbits;i++)
					{
					if(bt->nodes[i]) 
						{
						n = bt->nodes[i];
	
						if(n->expansion) n = n->expansion->parent;
						e = n->ext;
						if(e) { msi = e->msi; lsi = e->lsi; } else { msi = lsi = -1; }
						vt = vcdsav_splay(n, vt);
						if(!vt || vt->item != n)
							{
							unsigned char flags = 0;

							if(n->head.next)
							if(n->head.next->next)
								{
								flags = n->head.next->next->flags;
								}

							vt = vcdsav_insert(n, vt, ++nodecnt, flags, &n->head);
							}
						}
					}
				}
			}
		}

	if(!strace_append)
		{
		t=t->t_next;
		if(!t) 
			{
	                if(shadow_straces)
	                        {
	                        swap_strace_contexts();
				st = straces;

				strace_append = 1;
				t = st ? st->trace : NULL;
				}
			}
		}
		else
		{
		st = st->next;
		t = st ? st->trace : NULL;
		}
	}

if(strace_append) swap_strace_contexts();
if(!nodecnt) return(VCDSAV_EMPTY);


/* header */
if(lxt)
	{
	int dim;

	lt_set_chg_compress(lt);
	lt_set_clock_compress(lt);
	lt_set_initial_value(lt, 'x');	
	lt_set_time64(lt, 0);
	lt_symbol_bracket_stripping(lt, 1);

	switch(time_dimension)
		{
		case 'm':	dim = -3; break;
		case 'u':	dim = -6; break;
		case 'n':	dim = -9; break;
		case 'p':	dim = -12; break;
		case 'f':	dim = -15; break;
		default: 	dim = 0; break;
		}

	lt_set_timescale(lt, dim);
	}
	else
	{
	time(&walltime);
	fprintf(f, "$date\n");
	fprintf(f, "\t%s",asctime(localtime(&walltime)));
	fprintf(f, "$end\n");
	fprintf(f, "$version\n\t"WAVE_VERSION_INFO"\n$end\n");
	fprintf(f, "$timescale\n\t%d%cs\n$end\n", (int)time_scale, time_dimension);
	}

/* write out netnames here ... */
hp_clone = hp = calloc_2(nodecnt, sizeof(vcdsav_Tree *));
recurse_build(vt, &hp_clone);

for(i=0;i<nodecnt;i++)
	{
	char *netname = lxt ? hp[i]->item->nname : output_hier(hp[i]->item->nname);

	if(hp[i]->flags & (HIST_REAL|HIST_STRING))
		{
		if(lxt)
			{
			hp[i]->handle.p = lt_symbol_add(lt, netname, 0, 0, 0, hp[i]->flags & HIST_STRING ? LT_SYM_F_STRING : LT_SYM_F_DOUBLE);
			}
			else
			{
			fprintf(f, "$var real 1 %s %s $end\n", vcdid(hp[i]->val), netname);
			}
		}
		else
		{
		int msi = -1, lsi = -1;

		if(hp[i]->item->ext)
			{
			msi = hp[i]->item->ext->msi;
			lsi = hp[i]->item->ext->lsi;
			}
		
		if(msi==lsi)
			{
			if(lxt)
				{
				int strand_idx = strand_pnt(netname);
				if(strand_idx >= 0)
					{
					msi = lsi = atoi(netname + strand_idx + 1);
					}
				hp[i]->handle.p = lt_symbol_add(lt, netname, 0, msi, lsi, LT_SYM_F_BITS);
				}
				else
				{
				fprintf(f, "$var wire 1 %s %s $end\n", vcdid(hp[i]->val), netname);
				}
			}
			else
			{
			int len = (msi < lsi) ? (lsi - msi + 1) : (msi - lsi + 1);
			if(lxt)
				{
				hp[i]->handle.p = lt_symbol_add(lt, netname, 0, msi, lsi, LT_SYM_F_BITS);
				}
				else
				{
				fprintf(f, "$var wire %d %s %s $end\n", len, vcdid(hp[i]->val), netname);
				}
			hp[i]->len = len;
			if(len > max_len) max_len = len;
			}
		}

	}

row_data = malloc_2(max_len + 1);

if(!lxt)
	{
	output_hier("");
	free_hier();

	fprintf(f, "$enddefinitions $end\n");
	fprintf(f, "$dumpvars\n");
	}

/* value changes */

for(i=(nodecnt/2-1);i>0;i--)        /* build nodes into having heap property */
        {
        heapify(i,nodecnt);  
        }

for(;;)
	{
	heapify(0, nodecnt);

	if(!hp[0]->hist) break;
	if(hp[0]->hist->time > max_time) break;

	if((hp[0]->hist->time != prevtime) && (hp[0]->hist->time >= LLDescriptor(0)))
		{
		TimeType tnorm = hp[0]->hist->time;
		if(time_scale != 1)
			{
			tnorm /= time_scale;
			}

		if(lxt)
			{
			lt_set_time64(lt, tnorm);
			}
			else
			{
			fprintf(f, "#"TTFormat"\n", tnorm);
			}
		prevtime = hp[0]->hist->time;
		}
	
	if(hp[0]->hist->time >= LLDescriptor(0))
		{
		if(hp[0]->flags & (HIST_REAL|HIST_STRING))
			{
			if(hp[0]->flags & HIST_STRING)
				{
				char *vec = hp[0]->hist->v.h_vector ? hp[0]->hist->v.h_vector : "UNDEF";

				if(lxt)
					{
					lt_emit_value_string(lt, hp[0]->handle.p, 0, vec);
					}
					else
					{
					fprintf(f, "s%s %s\n", vec, vcdid(hp[0]->val));
					}
				}
				else
				{
				double *d = (double *)hp[0]->hist->v.h_vector;
                                double value;

				if(!d)
					{
	                                sscanf("NaN", "%lg", &value);
					}
					else
					{
					value = *d;
					}

				if(lxt)
					{
					lt_emit_value_double(lt, hp[0]->handle.p, 0, value);
					}
					else
					{
					fprintf(f, "r%.16g %s\n", value, vcdid(hp[0]->val));	
					}
				}	
			}
		else
		if(hp[0]->len)
			{
			if(hp[0]->hist->v.h_vector)
				{
				for(i=0;i<hp[0]->len;i++)
					{
					row_data[i] = analyzer_demang(lxt, hp[0]->hist->v.h_vector[i]);
					}				
				}
				else
				{
				for(i=0;i<hp[0]->len;i++)
					{
					row_data[i] = 'x';
					}				
				}
			row_data[i] = 0;
			
			if(lxt)
				{
				lt_emit_value_bit_string(lt, hp[0]->handle.p, 0, row_data);
				}
				else
				{
				fprintf(f, "b%s %s\n", vcd_truncate_bitvec(row_data), vcdid(hp[0]->val));
				}
			}
		else
			{
			if(lxt)
				{
				row_data[0] = analyzer_demang(lxt, hp[0]->hist->v.h_val);
				row_data[1] = 0;

				lt_emit_value_bit_string(lt, hp[0]->handle.p, 0, row_data);
				}
				else
				{
				fprintf(f, "%c%s\n", analyzer_demang(lxt, hp[0]->hist->v.h_val), vcdid(hp[0]->val));
				}
			}
		}

	hp[0]->hist = hp[0]->hist->next;
	}

if(prevtime < max_time)
	{
	if(lxt)
		{
		lt_set_time64(lt, max_time / time_scale);
		}
		else
		{
		fprintf(f, "#"TTFormat"\n", max_time / time_scale);
		}
	}


for(i=0;i<nodecnt;i++)
	{
	free_2(hp[i]);
	}

free_2(hp); hp = NULL;
free_2(row_data); row_data = NULL;

if(lxt)
	{
	lt_close(lt); lt = NULL;
	}
	else
	{
	fclose(f); f = NULL;
	}

return(VCDSAV_OK);
}

/************************ scopenav ************************/

struct namehier
{
struct namehier *next;
char *name;
char not_final;
};

static struct namehier *nhold=NULL;


void free_hier(void)
{
struct namehier *nhtemp;

while(nhold)
	{
	nhtemp=nhold->next;	
	free_2(nhold->name);
	free_2(nhold);
	nhold=nhtemp;
	}
}

/*
 * navigate up and down the scope hierarchy and
 * emit the appropriate vcd scope primitives
 */
static void diff_hier(struct namehier *nh1, struct namehier *nh2)
{
struct namehier *nhtemp;

if(!nh2)
	{
	while((nh1)&&(nh1->not_final))
		{
		fprintf(f, "$scope module %s $end\n", nh1->name);
		nh1=nh1->next;
		}
	return;
	}

for(;;)
	{
	if((nh1->not_final==0)&&(nh2->not_final==0)) /* both are equal */
		{
		break;
		}

	if(nh2->not_final==0)	/* old hier is shorter */
		{
		nhtemp=nh1;
		while((nh1)&&(nh1->not_final))
			{
			fprintf(f, "$scope module %s $end\n", nh1->name);
			nh1=nh1->next;
			}
		break;
		}

	if(nh1->not_final==0)	/* new hier is shorter */
		{
		nhtemp=nh2;
		while((nh2)&&(nh2->not_final))
			{
			fprintf(f, "$upscope $end\n");
			nh2=nh2->next;
			}
		break;
		}

	if(strcmp(nh1->name, nh2->name))
		{
		nhtemp=nh2;				/* prune old hier */
		while((nh2)&&(nh2->not_final))
			{
			fprintf(f, "$upscope $end\n");
			nh2=nh2->next;
			}

		nhtemp=nh1;				/* add new hier */
		while((nh1)&&(nh1->not_final))
			{
			fprintf(f, "$scope module %s $end\n", nh1->name);
			nh1=nh1->next;
			}
		break;
		}

	nh1=nh1->next;
	nh2=nh2->next;
	}
}


/*
 * output scopedata for a given name if needed, return pointer to name string
 */
char *output_hier(char *name)
{
char *pnt, *pnt2;
char *s;
int len;
struct namehier *nh_head=NULL, *nh_curr=NULL, *nhtemp;

pnt=pnt2=name;

for(;;)
{
while((*pnt2!='.')&&(*pnt2)) pnt2++;
s=(char *)calloc_2(1,(len=pnt2-pnt)+1);
memcpy(s, pnt, len);
nhtemp=(struct namehier *)calloc_2(1,sizeof(struct namehier));
nhtemp->name=s;

if(!nh_curr)
	{
	nh_head=nh_curr=nhtemp;
	}
	else
	{
	nh_curr->next=nhtemp;
	nh_curr->not_final=1;
	nh_curr=nhtemp;
	}

if(!*pnt2) break;
pnt=(++pnt2);
}

diff_hier(nh_head, nhold);
free_hier();
nhold=nh_head;

return(nh_curr->name);
}

